
;ling long kai fang unsigned 64-bit test

        include \masm32\include\masm32rt.inc
        .586
        include \masm32\macros\timers.asm

uArb2StrL9 PROTO :DWORD,:DWORD,:DWORD,:DWORD

LoopCount = 5000000

;--------------------------------------------------------------------------------------------

        .data

        ALIGN   4
InpVal  dq -1

        ALIGN   4
AscBuf  db      21 dup(?)

;--------------------------------------------------------------------------------------------

        .code

_main   proc

;restrict execution to a single core

        INVOKE  GetCurrentProcess
        INVOKE  SetProcessAffinityMask,eax,1

;run the test

        INVOKE  Sleep,500
        mov     ecx,5

loop00: push    ecx

        counter_begin LoopCount,HIGH_PRIORITY_CLASS
        INVOKE  uArb2StrL9,offset InpVal,8,offset AscBuf,21
        counter_end
        print   ustr$(eax)," clock cycles",13,10

        pop     ecx
        dec     ecx
        jnz     loop00

;exit

        inkey
        exit

_main   endp

;############################################################################################

        OPTION  EPILOGUE:NONE
        OPTION  PROLOGUE:NONE
        ALIGN   16

uArb2StrL9 PROC lpBinary:DWORD,dwNumberOfBytes:DWORD,lpBuffer:DWORD,dwSize:DWORD
;
; CONVERT UNSIGNED ARBITRARY INTEGER TO ASCII DECIMAL STRING
; LING LONG KAI FANG METHOD (HORNER'S RULE)
; BASE 4,294,967,296 to BASE 1,000,000,000 - 9 DIGITS PER DWORD - VERSION 1
; non-FPU/MMX/SSE - nondestructive
;
; Written by DednDave, November 2009
; Specialized code assistance by Drizz
; Tips from Steve Hutchinson, Raymond Filiatreault, Jochen (JJ2007), and qWord
; Additional tips from several other MASM32 forum members - http://www.masm32.com
;
;#########################################################
;
;Usage:
;
;       INVOKE  uArb2StrL9,
;               lpBinary,
;               dwNumberOfBytes,
;               lpBuffer,
;               dwSize
;
;        lpBinary = Offset of binary input value
; dwNumberOfBytes = Number of bytes in input value
;        lpBuffer = Offset of output buffer
;          dwSize = Size of output buffer in bytes
;
;-----------------------------------------
;
;Returns:
;
; edx = Pointer to first byte of output string (in this version, same as lpBuffer)
; ecx = Length of output string, in bytes
; eax = Return status:
;   eax = 0   ERROR_SUCCESS
;             "The operation completed successfully."
;          a) no errors:
;               the ASCII decimal string pointer and length are returned
;   eax = 87  ERROR_INVALID_PARAMETER
;             "The parameter is incorrect."
;          a) dwNumberOfBytes is 0, negative, or too large:
;               the specified output buffer pointer with length = 0 is returned
;          b) dwSize is less than 4 bytes
;               the specified output buffer pointer with length = 0 is returned
;   eax = 122 ERROR_INSUFFICIENT_BUFFER
;             "The data area passed to a system call is too small."
;             (windows list of error descriptions is poorly devised)
;          a) output buffer overflow:
;               the specified output buffer pointer with length = 0 is returned
;
;-----------------------------------------
;
;Notes:
;
; 1) Although the routine has a 1 Gb maximum integer size limit, it is recommended
;    that they are kept under 1 Mb. Larger integers may take a long time to evaluate.
;
; 2) The minimum output buffer size is 4 bytes.
;
; 3) Normally, the input value is not altered during operation. However, if the input
;    value need not be retained in binary form, the output string may overwrite the
;    input value, provided the input value is located at the lower address end of the
;    buffer and the buffer is large enough to accommodate the resulting string.
;
;    The maximum output string length may be calculated by:
;
;    L = Trunc( B * Log(2) ) + 1
;    where L is the string length in bytes and B is the integer length in bits
;
;    (Thank you Raymond and Drizz for help with the formula)
;
; 4) To achieve optimal performance, alignment of the input value and output buffer
;    is critical. They should both be 4-aligned. An additional slight increase in
;    performance may be obtained by using an output buffer a few bytes larger than
;    required. This avoids single-byte placement at the end of the string.
;
; 5) The output string is left-justified in the output buffer field. The buffer
;    address is returned in EDX and the length is returned in ECX for compatibility
;    with right-justified routines.
;
; 6) If output buffer size allows, the string is zero-terminated. The termination
;    byte is not counted in the returned length value.
;
; 7) No verification is made to insure the address parameters are valid (i.e. access
;    privileges). However, if adding the output buffer address to its' size causes
;    overflow, the parameter is flagged as invalid. Similarly, if adding the input
;    address to the input length causes overflow, the parameter is flagged as invalid.
;
;#########################################################
;
; Stack Frame Equates
;
;#########################################################

_OutBufSize   TEXTEQU <[EBP+32]> ;output buffer size in bytes
_OutBufBase   TEXTEQU <[EBP+28]> ;output buffer base address
_InpValSize   TEXTEQU <[EBP+24]> ;input value size in bytes
_InpValBase   TEXTEQU <[EBP+20]> ;input value base address

;                      [EBP+16]   PROC return address
;                      [EBP+12]   saved ESI contents
;                      [EBP+8]    saved EDI contents
;                      [EBP+4]    saved EBX contents
;                      [EBP]      saved EBP contents

_OutLastDword TEXTEQU <[EBP-4]>  ;address of last dword in output buffer

;another variable, _ActiveIndex, is defined in the uninitialized data area
;(see end of routine for the define)

;#########################################################
;
; Routine Initialization
;
;#########################################################

;establish the stack frame

        push    esi                     ;save the registers
        push    edi
        push    ebx
        push    ebp
        mov     ebp,esp                 ;stack frame base pointer

;validate input parameters

;EBP = stack frame base pointer

        mov     esi,_InpValSize         ;input value size
        cmp     esi,40000000h           ;maximum input size = 1 Gb
        ja      Error0                  ;error - invalid parameter

        or      esi,esi                 ;error if input size = 0
        mov     ecx,_InpValBase         ;input value base address
        jz      Error0                  ;error - invalid parameter

        dec     esi                     ;count to last byte
        add     esi,ecx                 ;address of last input byte
        jc      Error0                  ;error - invalid parameter

        inc     esi                     ;point to byte above input value
        mov     eax,_OutBufSize         ;EAX = output buffer size
        mov     ebx,3                   ;EBX = multipurpose 3
        lea     edi,[eax-1]             ;EDI = count to last byte
        cmp     eax,ebx                 ;minimum buffer size is 4 bytes
        jbe     Error0                  ;error - invalid parameter

        add     edi,_OutBufBase         ;last output buffer byte address
        jc      Error0                  ;error - invalid parameter

;4-align the last dword in the output buffer

        sub     edi,ebx                 ;last un-aligned output dword address
        test    edi,ebx                 ;is it 4-aligned ?
        jnz     Align1                  ;no - try to align it

Align0: mov     edx,ebx                 ;EDX = EBX = 3
        push    edi                     ;[ebp-4] = last output dword address
        xor     eax,eax                 ;EAX = 0

;scan to prevent evaluation of "unused" high order bytes

;Changing the direction flag is slow on some processors, so we "manually" scan down
;to the highest order meaningful byte. We start by scanning bytes until the address
;is 4-aligned, then scan dwords. It takes roughly 1 clock cycle per unused byte.

;EAX = 0
;EBX = 3
;ECX = input value base address
;EDX = 3
;ESI = address of byte above input value
;EBP = stack frame base pointer

UScan0: test    esi,ebx                 ;are we 4-aligned ?
        jz      UScan1                  ;yes - scan dwords

        dec     esi                     ;no - scan next byte
        cmp     esi,ecx                 ;base address ?
        jz      LCalc0                  ;yes - done - calculate length

        cmp     al,[esi]                ;is the byte meanignful ?
        jz      UScan0                  ;no - next byte

        jmp short LCalc0                ;yes - done - calculate length

;4-align the last dword in the output buffer (cont.)

;If the integer is 10 decimal digits or less, there are cases where
;aligning the data may cause buffer overwrite. We avoid this problem
;by testing the buffer size. If it is 10 bytes or less, the integer is
;small and mis-alignment does not have a signifigant effect on speed.

;EAX = output buffer length
;EDI = address of last unaligned output dword

Align1: cmp     eax,11                  ;is buffer length < 11 ?
        jb      Align0                  ;yes - don't align it

        and     edi,-4                  ;no - align the last dword
        jmp     Align0                  ;done - scan unused bytes

;-----------------------------------------

;invalid parameter error

;EBP = stack frame base pointer

Error0: push    87                      ;ERROR_INVALID_PARAMETER

;error exit

;EBP = stack frame base pointer
;return status code on the stack

Error1: mov     edx,_OutBufBase         ;buffer base address
        xor     ecx,ecx                 ;length = 0
        pop     eax                     ;return status code
        jmp     RExit0                  ;exit the routine

;buffer overflow error

;EBP = stack frame base pointer

Error2: push    122                     ;ERROR_INSUFFICIENT_BUFFER
        jmp     Error1                  ;error exit

;-----------------------------------------

;16-align UScan2

        db      6 dup(90h)

;scan to prevent evaluation of "unused" high order bytes (cont.)

UScan1: add     ebx,ecx                 ;EBX = base address + 3
        inc     edx                     ;EDX = 4

        ALIGN   16                      ;align the loop

UScan2: cmp     esi,ebx                 ;last dword ?
        jbe     UScan3                  ;yes - scan bytes

        sub     esi,edx                 ;no - next dword
        cmp     eax,[esi]               ;is the dword meaningful ?
        jz      UScan2                  ;no - next dword

        add     esi,edx                 ;yes - see which byte

UScan3: cmp     esi,ecx                 ;base address ?
        jz      LCalc0                  ;yes - done - calculate length

        dec     esi                     ;no - scan next byte
        cmp     al,[esi]                ;is the byte meanignful ?
        jz      UScan3                  ;no - next byte

;calculate minimum output buffer length

;We can pre-calculate the minimum number of bytes required in the output buffer,
;which means we don't have to test for overflow during base conversion. This may
;actually be slower for short integers, as it takes ~25 cycles to calculate.

;L = Trunc( (B - 1) * Log(2) ) + 1
;where L is the string length in bytes and B is the integer length in bits

;We use the low order bit of EBX to store the high order bit of the bit count, and recover
;it at the end with an SHR instruction and use ADD/ADC to "multiply" that bit.

;EAX = 0
;ECX = input value base address
;ESI = address of highest order meaningful input byte
;EBP = stack frame base pointer

LCalc0: push    esi                     ;save highest input byte address
        mov     al,[esi]                ;get the byte for bit count
        sub     esi,ecx                 ;byte count under highest byte
        or      eax,eax                 ;
        push    esi                     ;save byte count under highest byte
        jz      LCalc1                  ;only 0 if value is 0 or -1

        bsr     ax,ax                   ;bit count of high order byte

LCalc1: shl     esi,3                   ;bits = 8 * bytes
        rcl     ebx,1                   ;it can be a 33 bit value
        add     eax,esi                 ;add the bits from high byte
        mov     ecx,7DE7FBCCh           ;multiplier lo: 2^64 * Log(2)
        mov     edi,eax                 ;save the bit count lo in EDI
        mul     ecx                     ;partial multiply
        mov     eax,4D104D42h           ;multiplier hi: 2^64 * Log(2)
        xchg    eax,edi                 ;recall the bit count lo
        mov     esi,edx                 ;save the partial
        mul     edi                     ;partial multiply
        shr     ebx,1                   ;recover the high order bit
        jc      LCalc3                  ;single bit "multiply" by addition

LCalc2: xor     ecx,ecx                 ;zero for carry
        add     eax,esi                 ;complete the product
        adc     edx,ecx                 ;EDX = minimum length - 1
        mov     eax,ecx                 ;EAX = ECX = 0
        cmp     edx,_OutBufSize         ;do we have enough buffer ?
        mov     cl,3                    ;ECX = 3 for the aligned load
        jae     Error2                  ;no - output buffer overflow error

;align and load the first (high order) input dword

;EAX = 0
;ECX = 3
;EBP = stack frame base pointer

        pop     edx                     ;recall byte count under highest byte
        and     ecx,edx                 ;ECX = offset correction
        pop     esi                     ;recall highest input byte address
        shr     edx,1                   ;if bit 0 = 0, we load 1 or 3 bytes
        mov     edi,_OutLastDword       ;EDI = last output dword address
        jnc     DLoad0                  ;if bit 0 = 1, we load 2 or 4 bytes

        sub     esi,ecx                 ;apply offset correction
        shr     edx,1                   ;if bit 1 = 0, we load 2 bytes
        jnc     DLoad1                  ;if bit 1 = 1, we load 4 bytes

        mov     eax,[esi]               ;load 4 bytes
        jmp short DLoad2                ;set up for base conversion

;single bit "multiply" by addition

LCalc3: add     eax,ecx                 ;add the multiplier lo
        adc     edx,edi                 ;and the multiplier hi
        jmp     LCalc2                  ;complete the calculation

;16-align Llkf90

        db      7 dup(90h)

;align and load the first (high order) input dword (cont.)

DLoad0: mov     al,[esi]                ;if offset correction = 0, we load 1 byte
        jecxz   DLoad2                  ;otherwise, we load 3 bytes

        rol     eax,16                  ;reorder the bytes
        sub     esi,ecx                 ;apply offset correction

DLoad1: mov     ax,[esi]                ;load 2 bytes

DLoad2: mov     _ActiveIndex,edi        ;initialize the active dword index
        mov     cl,ch                   ;ECX = 0
        cmp     esi,_InpValBase         ;is this the only input dword ?
        jz      Llkf96                  ;yes - do last input dword

;#########################################################
;
; Ling Long Kai Fang Base Conversion
;
; Convert base 4,294,967,296 to base 1,000,000,000
;
;#########################################################

;The outer part of the loop handles the addition and multiplication of a new input
;dword to the first dword in the output accumulator. The low order accumulator dword
;may be as large as 999,999,744 (before addition of the last input dword). The input
;dword may be as large as 4,294,969,295. We want to add those two together, multiply
;by 4,294,967,296, then divide by 1,000,000,000. This creates a 65-bit dividend. But,
;because the lower 9 bits of the remainder are always 0, we can reduce the dividend
;by as many as 9 bits. To simplify the code, we reduce it by 8 bits. So, we instead
;multiply by 16,777,216, then divide by 3,906,250. This multiplication is performed
;by rearranging the input data bytes. The resulting carry into the inner loop will
;never be more than 35 bits, so the inner loop multiply/divide does not have to
;handle values as large. It does, however, have to handle a full 30 bit remainder.

        xchg    eax,ecx                 ;swap for 16m multiplication

;ling long kai fang outer loop

        ALIGN   16                      ;align the loop

Llkf90: xchg    al,cl                   ;swap for 16m multiplication
        push    esi                     ;save input pointer
        ror     eax,8                   ;reorder for 16m multiplication
        push    ebp                     ;save stack frame base pointer
        mov     edx,0A09B5A53h          ;multiplier lo: 2^76 / 3906250
        mov     ebp,eax                 ;save lo dword for remainder
        push    edx                     ;save multiplier lo on stack
        ror     ecx,8                   ;reorder for 16m multiplication
        mul     edx                     ;partial multiply to divide
        mov     eax,44B82Fh             ;multiplier hi: 2^76 / 3906250
        mov     esi,edx                 ;partial product
        mul     ebp                     ;partial multiply to divide
        xor     ebx,ebx                 ;zero for carry
        add     esi,eax                 ;partial product
        pop     eax                     ;recall multiplier lo
        adc     ebx,edx                 ;partial product
        mul     ecx                     ;partial multiply to divide
        add     esi,eax                 ;partial product
        adc     ebx,edx                 ;partial product
        mov     eax,44B82Fh             ;multiplier hi: 2^76 / 3906250
        mul     ecx                     ;partial multiply to divide
        xor     ecx,ecx                 ;zero for carry
        add     eax,ebx                 ;partial product
        adc     ecx,edx                 ;partial product
        shrd    eax,ecx,12              ;adjust quotient lo (carry)
        shr     ecx,12                  ;adjust quotient hi (carry)
        imul    edx,eax,3906250         ;1,000,000,000 / 256
        sub     ebp,edx                 ;subtract for remainder
        shl     ebp,8                   ;adjust the remainder
        mov     [edi],ebp               ;store remainder in accumulator
        sub     edi,4                   ;next accumulator dword
        cmp     edi,_ActiveIndex        ;is it active yet ?
        jb      Llkf93                  ;no - activate it

;ling long kai fang inner loop

Llkf91: add     ecx,[edi]               ;add the accumulator dword to carry

Llkf92: mov     ebp,eax                 ;save lo dword for remainder
        mov     edx,26D694B3h           ;multiplier lo: 2^90 / 1,000,000,000
        xor     ebx,ebx                 ;zero for carry
        mul     edx                     ;partial multiply to divide
        mov     eax,112E0BE8h           ;multiplier hi: 2^90 / 1,000,000,000
        push    eax                     ;save multiplier hi on stack
        mov     esi,edx                 ;partial product
        mul     ebp                     ;partial multiply to divide
        add     esi,eax                 ;partial product
        mov     eax,26D694B3h           ;multiplier lo: 2^90 / 1,000,000,000
        adc     ebx,edx                 ;partial product
        mul     ecx                     ;partial multiply to divide
        add     esi,eax                 ;partial product
        pop     eax                     ;recall multiplier hi
        adc     ebx,edx                 ;partial product
        mul     ecx                     ;partial multiply to divide
        xor     ecx,ecx                 ;zero for carry
        add     eax,ebx                 ;partial product
        adc     ecx,edx                 ;partial product
        shrd    eax,ecx,26              ;adjust quotient lo (carry)
        imul    edx,eax,1000000000      ;1,000,000,000
        shr     ecx,26                  ;adjust quotient hi (carry)
        sub     ebp,edx                 ;subtract for remainder
        mov     [edi],ebp               ;store remainder in accumulator
        sub     edi,4                   ;point to next accumulator dword
        cmp     edi,_ActiveIndex        ;is it active yet ?
        jae     Llkf91                  ;yes - next accumulator dword

Llkf93: or      eax,eax                 ;no - any carry lo ?
        jz      Llkf95                  ;no - test carry hi

Llkf94: mov     _ActiveIndex,edi        ;yes - activate it
        jmp     Llkf92                  ;next accumulator dword

Llkf95: or      ecx,ecx                 ;any carry hi ?
        jnz     Llkf94                  ;yes - activate it

;continuation of outer loop

        pop     ebp                     ;recall stack frame base pointer
        pop     esi                     ;recall input pointer
        mov     edi,_OutLastDword       ;point to low order accumulator
        sub     esi,4                   ;point to next input dword
        mov     ecx,[edi]               ;load the accumulator dword
        add     ecx,[esi]               ;add the new input dword
        adc     eax,eax                 ;allow it to carry
        cmp     esi,_InpValBase         ;is this the last input dword ?
        ja      Llkf90                  ;no - loop until it is

        xchg    eax,ecx                 ;swap for 16m multiplication

;last input dword

;The last input dword does not get multiplied, so it has its' own simplified code.
;In this case, it may be a 33 bit dividend. We save the low order bit, divide as a
;32 bit value, then recover the saved bit into the remainder. Any carry that may
;be generated is handled by a very simple and fast "subtract to divide" loop.

Llkf96: shr     ecx,1                   ;lo bit of hi dword into lo dword
        mov     ebx,44B82FA1h           ;multiplier: 2^59 / 500,000,000
        rcr     eax,1                   ;lo bit of hi dword into lo dword
        rcl     ecx,1                   ;save lo bit of lo dword
        mov     esi,eax                 ;save lo dword for remainder
        mul     ebx                     ;multiply to divide
        shr     edx,27                  ;adjust quotient (carry)
        imul    eax,edx,500000000       ;1,000,000,000 / 2
        sub     esi,eax                 ;subtract for remainder
        shr     ecx,1                   ;recall lo dword lo bit
        rcl     esi,1                   ;restore lo bit into remainder
        mov     [edi],esi               ;store remainder in accumulator
        or      edx,edx                 ;any carry ?
        mov     esi,_ActiveIndex        ;ESI = active index
        jz      Ascii0                  ;no - convert to ASCII

        mov     ebx,4                   ;yes - EBX = 4 for subtract to divide
        inc     ecx                     ;ECX = 1
        mov     eax,1000000000          ;EAX = 1,000,000,000
        jmp short Llkf98                ;subtract to divide loop entry

;last input dword loop

        ALIGN   16                      ;align the loop

Llkf97: add     edx,[edi]               ;add accumulator to carry
        cmp     edx,eax                 ;carry into next dword ?
        jb      Llkf99                  ;no - done - store accumulator

        sub     edx,eax                 ;subtract to divide
        mov     [edi],edx               ;store remainder in accumulator
        mov     edx,ecx                 ;carry is always 1

Llkf98: sub     edi,ebx                 ;point to next accumulator dword
        cmp     edi,esi                 ;is it active ?
        jae     Llkf97                  ;yes - test carry until done

        mov     esi,edi                 ;no - update the active index

Llkf99: mov     [edi],edx               ;store last accumulator dword

;#########################################################
;
; ASCII Conversion
;
;#########################################################

;determine the number of digits in the first dword

;ECX = 0 or 1
;ESI = active index pointer (offset of highest order accumulator dword)
;EBP = stack frame base pointer

Ascii0: mov     cl,9                    ;ECX = 9
        mov     eax,[esi]               ;load the first dword
        mov     bl,3                    ;BL = 3
        push    eax                     ;save the first dword

;scale the first dword by 1,000 to count digits
;we may make 0, 1, or 2 passes on this loop

Ascii1: cmp     eax,1000000             ;is it  >= 1,000,000 ?
        jae     Ascii2                  ;yes - determine mod 3 count

        sub     cl,bl                   ;no - subtract 3 from count
        imul    eax,eax,1000            ;scale it by 1,000
        cmp     cl,bl                   ;last 3 digits ?
        ja      Ascii1                  ;no - until it is

;once the binary value has been scaled, we determine the mod 3 count

Ascii2: cmp     eax,100000000           ;is it >= 100,000,000 ?
        jae     Ascii3                  ;yes - calculate length

        dec     ecx                     ;adjust mod 3 count
        cmp     eax,10000000            ;no - is it >= 10,000,000 ?
        sbb     cl,ch                   ;adjust mod 3 count with CF

;calculate the total string length and verify sufficient output buffer space for the string

Ascii3: mov     edx,_OutLastDword       ;low order accumulator dword address
        sub     edx,esi                 ;EDX = 4 * (accumulator dwords - 1)
        shr     edx,2                   ;EDX = accumulator dwords - 1
        pop     eax                     ;recall the first accumulator dword
        imul    edi,edx,9               ;9 digits per dword
        add     edi,ecx                 ;add the first dword digit count
        cmp     edi,_OutBufSize         ;do we have enough output buffer ?
        ja      Error2                  ;error - output buffer overflow

;determine the multiplier table vector

        dec     ecx                     ;ECX = first dword digits - 1
        push    ebp                     ;save stack frame base pointer
        inc     edx                     ;EDX = accumulator dwords
        and     ebx,ecx                 ;EBX = (first dword digits - 1) mod 3
        mov     edi,_OutBufBase         ;EDI = output buffer base address
        imul    ebx,ebx,28              ;EBX = multiplier table index

;set up the stack values and determine the loop entry point

        xor     ebp,ebp                 ;EBP = 0 (previous pass digits binary)
        push    edx                     ;accumulator dword count on stack for loop
        add     ebx,offset _AscTable    ;EBX = multiplier table vector
        shr     ecx,2                   ;first dword = 1 to 4 digits ?
        push    esi                     ;accumulator pointer on stack for loop
        jz      Ascii7                  ;yes - loop entry

        dec     ecx                     ;first dword = 5 to 8 digits ?
        jnz     Ascii5                  ;yes - loop entry

        jmp short Ascii6                ;no - 8 to 10 digits loop entry

;ASCII conversion loop

;This loop converts the accumulator 9-digit binary dwords into an ASCII string.
;It accesses memory by 4-aligned dwords only. I like to call it a "merry-go-round"
;loop because it has multiple entry points and is easy to ride or get off, but hard
;to know how and where to get on. I was pleasantly surprised when it worked the first
;time I tried it. I was further surprised to find it wasn't that much slower than a
;simpler 8 digit loop (it just takes a lot more code). The entry point is determined
;by the number of digits in the first dword. There are 3 basic sections to the loop.
;Each section is used three different ways, depending on digit alignment in dwords
;(a mod 9 function). The EBX register holds the multiplier table vector (a mod 4
;function). The table vector varies in each section with each loop pass. This 9 and
;4 relationship synchronizes the 9 digits per accumulator dword with the 4 digits
;placed in each ASCII string dword.

        ALIGN   16                      ;align the loop

Ascii4: push    ecx                     ;save accumulator dword count
        mov     eax,[esi]               ;load the next dword
        push    esi                     ;save accumulator source index

Ascii5: mov     edx,[ebx]               ;multiply to divide multiplier from table
        mov     esi,eax                 ;save the value for the remainder
        mul     edx                     ;multiply to divide
        mov     ecx,[ebx+4]             ;shift count from table
        mov     eax,4294968             ;multiplier: 2^32 / 1,000
        shr     edx,cl                  ;adjust quotient by variable count
        push    ebx                     ;save the table vector base
        xadd    edx,ebp                 ;add and swap the value/quotient
        mul     edx                     ;multiply to divide
        mov     ebx,10                  ;EBX = digit multiplier
        mov     ecx,edx                 ;digit 1 in CL
        mul     ebx                     ;multiply to divide
        mov     ch,dl                   ;digit 2 in CH
        mul     ebx                     ;multiply to divide
        bswap   ecx                     ;digits 1 & 2 up hi
        mov     ch,dl                   ;digit 3 in CH
        mul     ebx                     ;multiply to divide - digit 4 in DL
        pop     ebx                     ;recall the table vector base
        lea     ecx,[edx+ecx+'0000']    ;make them all ASCII
        mov     eax,[ebx+8]             ;remaining digits multiplier from table
        bswap   ecx                     ;reorder the digits
        mul     ebp                     ;multiply remaining digits
        mov     [edi],ecx               ;store 4 ASCII digits
        xchg    eax,esi                 ;swap remainder with original
        add     edi,4                   ;point to next string dword
        sub     eax,esi                 ;remaining digits for next section

Ascii6: mov     edx,[ebx+12]            ;multiply to divide multiplier from table
        mov     esi,eax                 ;save the value for the remainder
        mul     edx                     ;multiply to divide
        mov     ecx,[ebx+16]            ;shift count from table
        mov     eax,4294968             ;multiplier: 2^32 / 1,000
        shr     edx,cl                  ;adjust quotient by variable count
        push    ebx                     ;save the table vector base
        mov     ebp,edx                 ;save remaining digits binary
        mul     edx                     ;multiply to divide
        mov     ebx,10                  ;EBX = digit multiplier
        mov     ecx,edx                 ;digit 1 in CL
        mul     ebx                     ;multiply to divide
        mov     ch,dl                   ;digit 2 in CH
        mul     ebx                     ;multiply to divide
        bswap   ecx                     ;digits 1 & 2 up hi
        mov     ch,dl                   ;digit 3 in CH
        mul     ebx                     ;multiply to divide - digit 4 in DL
        pop     ebx                     ;recall the table vector base
        lea     ecx,[edx+ecx+'0000']    ;make them all ASCII
        mov     eax,[ebx+20]            ;remaining digits multiplier from table
        bswap   ecx                     ;reorder the digits
        mul     ebp                     ;multiply remaining digits
        mov     [edi],ecx               ;store 4 ASCII digits
        xchg    eax,esi                 ;swap remainder with original
        add     edi,4                   ;point to next string dword
        sub     eax,esi                 ;remaining digits for next section

Ascii7: cmp     ebx,offset _AscTable+84 ;is it the last row of the table array ?
        mov     edx,[ebx+24]            ;scaling multiplier from table
        jz      Ascii9                  ;yes - means there are 4 digits remaining

        add     ebx,28                  ;28 = array element row step
        mul     edx                     ;no - not enough digits to make a dword
        xchg    eax,ebp                 ;store the scaled remaining digits binary

Ascii8: pop     esi                     ;recall accumulator source index
        add     esi,4                   ;point to next accumulator dword
        pop     ecx                     ;recall accumulator dword count
        dec     ecx                     ;decrement accumulator dword count
        jnz     Ascii4                  ;until all accumulator dwords are done

        jmp short AsciiA                ;dword loop done - do few remaining digits

Ascii9: mov     ebp,10                  ;EBP = digit multiplier
        mul     edx                     ;multiply to divide
        mov     ecx,edx                 ;digit 1 in CL
        mul     ebp                     ;multiply to divide
        mov     ch,dl                   ;digit 2 in CH
        mul     ebp                     ;multiply to divide
        bswap   ecx                     ;digits 1 & 2 up hi
        mov     ch,dl                   ;digit 3 in CH
        mul     ebp                     ;multiply to divide
        lea     ecx,[edx+ecx+'0000']    ;make them all ASCII
        xor     ebp,ebp                 ;exactly 4 digits - no remaining binary
        bswap   ecx                     ;reorder the digits
        sub     ebx,84                  ;3 * 28 - restart table cycle
        mov     [edi],ecx               ;store 4 ASCII digits
        add     edi,4                   ;point to next string dword
        jmp     Ascii8                  ;resume conversion loop

;ASCII dword loop complete - see if there are any digits left

;EBX = current table vector
;ECX = 0
;EDI = output string pointer
;EBP = remaining digits binary

AsciiA: mov     esi,ebp                 ;ESI = remaining digits binary
        pop     ebp                     ;recall stack frame base pointer
        mov     eax,_OutBufBase         ;EAX = output buffer base pointer
        push    eax                     ;save output buffer base pointer
        add     eax,_OutBufSize         ;EAX = address of byte above buffer
        sub     ebx,offset _AscTable    ;subtract the table base address
        jz      AsciiC                  ;no digits left - do null terminator

;from the table vector, calculate the remaining digit count (1 to 3)

        push    eax                     ;save address of byte above buffer
        mov     eax,924924Ah            ;multiplier: 2^32 / 28
        mul     ebx                     ;multiply to divide
        mov     eax,4294968             ;multiplier: 2^32 / 1,000
        mov     ecx,edx                 ;ECX = remaining digit count

;convert the last few digits to ASCII

        mul     esi                     ;multiply to divide
        mov     esi,10                  ;ESI = digit multiplier
        mov     ebx,edx                 ;digit 1 in BL
        mul     esi                     ;multiply to divide
        mov     esi,655360              ;multiplier: 10 * 2^16
        mov     bh,dl                   ;digit 2 in BH
        mul     esi                     ;multiply to divide
        mov     dx,bx                   ;EDX = combined digits
        pop     eax                     ;recall address of byte above buffer
        or      edx,'000'               ;make them all ASCII

;determine if the buffer is large enough for dword placement

;EAX = offset of byte above output buffer
;ECX = remaining digit count (1 to 3)
;EDX = remaining ASCII digits
;EDI = output string pointer
;EBP = stack frame base pointer

        lea     ebx,[edi+4]             ;dword test address
        cmp     eax,ebx                 ;do we have room for a dword ?
        jae     AsciiD                  ;yes - write them as a dword

;the buffer is not large enough to allow dword placement
;of the last few digits, so we place each byte individually

AsciiB: mov     [edi],dl                ;place a single byte
        shr     edx,8                   ;next byte in DL
        inc     edi                     ;point to next string byte
        dec     ecx                     ;decrement the byte count
        jnz     AsciiB                  ;until all bytes are placed

;if buffer space allows, place a null string terminator

;EAX = offset of byte above output buffer
;ECX = 0
;EDI = output string pointer
;EBP = stack frame base pointer

AsciiC: cmp     eax,edi                 ;is there buffer space for a terminator ?
        jbe     AsciiG                  ;no - skip null terminator placement

        mov     [edi],cl                ;yes - place a null byte
        jmp short AsciiG                ;set the return values

;the buffer is large enough to allow dword placement of the last few digits
;mask off unwanted high order bytes, then place the dword
;the dword will include one or more null terminator byte(s)

;EAX = offset of byte above output buffer
;ECX = remaining digit count (1 to 3)
;EDX = remaining ASCII digits
;EDI = output string pointer
;EBP = stack frame base pointer

AsciiD: cmp     cl,2                    ;how many bytes are left ?
        ja      AsciiF                  ;3 bytes left - place all of them

        jz      AsciiE                  ;2 bytes left - zero high 16 bits first

        mov     dh,ch                   ;only 1 byte left - zero digit 2 also

AsciiE: movzx   edx,dx                  ;zero the high 16 bits

AsciiF: mov     [edi],edx               ;place last digits in the string
        add     edi,ecx                 ;adjust the string pointer

;set return value registers

;EDI = current output string pointer (address of terminator)
;EBP = stack frame base pointer

AsciiG: xor     eax,eax                 ;EAX = 0 for return status code
        pop     edx                     ;recall output buffer base pointer
        mov     ecx,edi                 ;ECX = string pointer
        sub     ecx,edx                 ;ECX = string length, less terminator

;#########################################################
;
; Routine Exit
;
;#########################################################

;EAX = return status code
;ECX = string length
;EDX = string address (output buffer base address)
;EBP = stack frame base pointer

RExit0: leave                           ;same as MOV ESP,EBP ; POP EBP
        pop     ebx                     ;restore the registers
        pop     edi
        pop     esi
        ret     16                      ;return and discard 4 dwords

;#########################################################
;
; ASCII Conversion Multiplier Table
;
;#########################################################

;This multiplier table is used by the ASCII conversion loop. The array has 4 rows, each
;consisting of 7 dword elements. The first 3 elements are similar to the second 3. They
;include a multiplier and a shift count for multiply-to-divide and a multiplier for the
;quotient-divisor product used to obtain the remainder. The 7th element in each row is
;a multiplier used to scale the remaining digits for the next loop pass whenever there
;are less than 4 digits of information. On loop passes where there are 4 remaining
;digits, it is the multiplier for multiply-to-divide conversion (2^32 / 1,000).

        ALIGN   4

_AscTable dd 53E2D624h,15,100000
          dd 0CCCCCCCDh,3,10
          dd 1000
          dd 431BDE83h,18,1000000
          dd 0A3D70A3Eh,6,100
          dd 100
          dd 35AFE536h,21,10000000
          dd 83126E98h,9,1000
          dd 10
          dd 2AF31DC5h,24,100000000
          dd 68DB8BADh,12,10000
          dd 4294968
;end of _AscTable

;#########################################################
;
; Uninitialized Data
;
;#########################################################

;The _ActiveIndex variable stores a pointer that represents the
;lowest address of the dwords in the accumulator (highest order
;dword) that are in use. It is located in uninitialized data
;because the EBP register is used for math during base conversion.

        .DATA?
        ALIGN   4

_ActiveIndex    dd ?

        .CODE

;#########################################################

uArb2StrL9 ENDP

        OPTION  EPILOGUE:EpilogueDef
        OPTION  PROLOGUE:PrologueDef

;############################################################################################

        end     _main

